from silence.settings import settings
from silence.cli.template_downloader import download_from_github

def handle(argv):
    project_name = argv[0] if len(argv) >= 1 else settings.PROJECT_TEMPLATE_NAME
    repo_url = argv[1] if len(argv) >= 2 else settings.PROJECT_TEMPLATE_REPO

    download_from_github(project_name, repo_url)

    print(f'The Silence project "{project_name}" has been created.')


from functools import wraps
from silence.exceptions import HTTPError, DatabaseError

import re

regex_error_str = re.compile(r"""\(.*?, ['"](.*)['"]\)""")

# Wraps a DB query/update call to catch any possible DatabaseErrors
# and wrap them inside a HTTPError
def db_call(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except DatabaseError as dberr:
            # Grab the error message from the exception
            m = regex_error_str.match(str(dberr))
            msg = m.group(1) if m else str(dberr)
            code = 400
            raise HTTPError(code, msg)
    return wrapper


from pymysql.cursors import DictCursor

from silence.db.connector import get_conn
from silence.exceptions import DatabaseError
from silence.decorators import db_call
from silence.logging.default_logger import logger

###############################################################################
# The DAL (Data Access Layer) functions provide an abstraction layer
# for querying and updating the database.
###############################################################################

# Query method to retrieve information
def query(q, params=None):
    logger.debug(f'Executing SQL query "{q}" with params {params}')

    # Fetch the connection and get a cursor
    conn = get_conn()
    cursor = conn.cursor(DictCursor)
    try:
        # Execute the query, with or without parameters and return the result
        if params:
            cursor.execute(q, params)
        else:
            cursor.execute(q)

        res = cursor.fetchall()
        logger.debug(f"Query result: {res}")
        return res
    except Exception as exc:
        # If anything happens, wrap the exceptions in a DatabaseError
        raise DatabaseError(exc) from exc
    finally:
        # Close the cursor and the connection
        cursor.close()
        conn.close()


# Update method to modify information
def update(q, params=None):
    logger.debug(f'Executing SQL operation "{q}" with params {params}')
    conn = get_conn()
    cursor = conn.cursor(DictCursor)

    # Fetch the connection and get a cursor
    try:
        # Execute the query, with or without parameters and return the result
        if params:
            cursor.execute(q, params)
        else:
            cursor.execute(q)

        conn.commit()

        # Return the ID of the row that was modified or inserted
        lastid = cursor.lastrowid
        logger.debug(f"Last modified row ID: {lastid}")
        res = {"lastId": lastid}
        return res
    except Exception as exc:
        # If anything happens, wrap the exceptions in a DatabaseError
        raise DatabaseError(exc) from exc
    finally:
        # Close the cursor and the connection
        cursor.close()
        conn.close()

###############################################################################
# Safe wrappers for the API, which return an HTTPError instead of a
# DatabaseError
###############################################################################

@db_call
def api_safe_query(*args, **kwargs):
    return query(*args, **kwargs)

@db_call
def api_safe_update(*args, **kwargs):
    return update(*args, **kwargs)



from flask import jsonify, request
from werkzeug.security import check_password_hash, generate_password_hash

from silence.auth.tokens import create_token, check_token
from silence.db import dal
from silence.sql.builder import get_login_query, get_register_user_query
from silence.sql.table_cols import get_table_cols
from silence.settings import settings
from silence.exceptions import HTTPError
from silence.logging.default_logger import logger
from silence.server import manager as server_manager

###############################################################################
# Defines the default endpoints provided by Silence,
# mainly /login and /register
###############################################################################

def show_api_endpoints():
    return jsonify(server_manager.API_TREE.get_endpoint_list()), 200

def login():
    USERS_TABLE, IDENTIFIER_FIELD, PASSWORD_FIELD = get_login_settings()
    # Ensure that the user has sent the required fields
    form = request.json if request.is_json else request.form
    form = filter_fields_db(form, USERS_TABLE)

    username = form.get(IDENTIFIER_FIELD, None)
    password = form.get(PASSWORD_FIELD, None)

    if not username or not password:
        raise HTTPError(400, f"Both '{IDENTIFIER_FIELD}' and '{PASSWORD_FIELD}' are required")

    logger.debug(f"Login request from user {username} with password {password}")

    # Look if there is an user with such username
    q = get_login_query(USERS_TABLE, IDENTIFIER_FIELD, username)
    users = dal.api_safe_query(q)

    if not users:
        logger.debug(f"The identifier {username} was not found")
        raise HTTPError(400, "User not found")

    # The identifier field should be unique (/register also takes care of that)
    # so we can just extract the first one
    user = users[0]
    
    # Check if the user's password matches the provided one
    if PASSWORD_FIELD not in user:
        raise HTTPError(500, f"The user has no attribute '{PASSWORD_FIELD}'")

    password_ok = check_password_hash(user[PASSWORD_FIELD], password)
    if not password_ok:
        logger.debug(f"Incorrect password")
        raise HTTPError(400, "The password is not correct")

    # If we've reached here the login is successful, generate a session token
    # and return it with the logged user's info
    logger.debug("Login OK")
    token = create_token(user)
    del user[PASSWORD_FIELD]
    res = {"sessionToken": token, "user": user}
    return jsonify(res), 200

def register():
    USERS_TABLE, IDENTIFIER_FIELD, PASSWORD_FIELD = get_login_settings()
    
    # Ensure that the user has sent the required fields
    form = request.json if request.is_json else request.form
    form = filter_fields_db(form, USERS_TABLE)

    username = form.get(IDENTIFIER_FIELD, None)
    password = form.get(PASSWORD_FIELD, None)

    if not username or not password:
        raise HTTPError(400, f"Both '{IDENTIFIER_FIELD}' and '{PASSWORD_FIELD}' are required")

    logger.debug(f"Register request with data {form}")

    # Ensure that the identifier is unique
    login_q = get_login_query(USERS_TABLE, IDENTIFIER_FIELD, username)
    other_users = dal.api_safe_query(login_q)

    if other_users:
        logger.debug(f"The identifier {username} already exists")
        raise HTTPError(400, f"There already exists another user with that {IDENTIFIER_FIELD}")

    # Create the user object, replacing the password with the hashed one
    user = dict(form)
    user[PASSWORD_FIELD] = generate_password_hash(password)

    # Try to insert it in the DB
    # Since the /register endpoint must adapt to any possible table,
    # we assume that the user knows what they're doing and submits the
    # appropriate fields. Otherwise, the DB will just complain.
    register_q = get_register_user_query(USERS_TABLE, user)
    dal.api_safe_update(register_q)

    # Fetch the newly created user from the DB (some fields may have been)
    # automatically generated, like its ID
    # It should now exist
    user = dal.api_safe_query(login_q)[0]

    # If we've reached here the register is successful, generate a session token
    # and return it with the logged user's info
    logger.debug("Register OK")
    token = create_token(user)
    del user[PASSWORD_FIELD]
    res = {"sessionToken": token, "user": user}
    return jsonify(res), 200


###############################################################################
# Aux functions

# Transforms the received dict of fields into a filtered one that shares
# the same capitalization with the DB columns
def filter_fields_db(data, table_name):
    cols = get_table_cols(table_name)
    res = {}

    for field, value in data.items():
        for col_name in cols:
            if field.lower() == col_name.lower():
                res[col_name] = value
    
    return res

# Returns a given column name with the correct capitalization for its table
# Raises a ValueError if it can't be found in the given table
def col_correct_case(col_name, table_name):
    cols = get_table_cols(table_name)

    for col in cols:
        if col.lower() == col_name.lower():
            return col

    raise ValueError(f"The column {col_name} could not be found in table {table_name}")


# Returns the login table and fields as specified in the settings,
# with the correct capitalization to avoid SQL errors
def get_login_settings():
    users_table = settings.USER_AUTH_DATA["table"]
    identifier_field = col_correct_case(settings.USER_AUTH_DATA["identifier"], users_table)
    password_field = col_correct_case(settings.USER_AUTH_DATA["password"], users_table)
    return users_table, identifier_field, password_field



from silence.db.dal import query

global TABLE_COLUMNS
TABLE_COLUMNS = {}

# Returns the list of names for the columns of a table, storing it
# after the first query for a given table
def get_table_cols(table_name):
    global TABLE_COLUMNS

    if table_name not in TABLE_COLUMNS:
        cols = query(f"SHOW COLUMNS FROM {table_name}")
        col_names = [col["Field"] for col in cols]
        TABLE_COLUMNS[table_name] = col_names
    
    return TABLE_COLUMNS[table_name]



# Thanks to https://stackoverflow.com/a/3885198/5604339 :)
import decimal
import json
from datetime import datetime

from silence.settings import settings

class SilenceJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            func = str if settings.DECIMALS_AS_STRINGS else float
            return func(o)

        try:
            return super().default(o)
        except TypeError:
            return str(o)

class SilenceJSONSerializer:
    def dumps(o):
        return json.dumps(o, cls=SilenceJSONEncoder)

    def loads(s):
        return json.loads(s)


from silence.server import manager as server_manager
from silence.logging.default_logger import logger
from silence.settings import settings
from silence.utils.check_update import check_for_new_version
from silence import __version__


def handle(argv):
    logger.info(f"Silence v{__version__}")
    logger.debug("Current settings:\n" + str(settings))

    new_ver = check_for_new_version()
    if new_ver:
        logger.warning(f"A new Silence version (v{new_ver}) is available. Run 'pip install --upgrade Silence' to upgrade.")

    server_manager.setup()
    server_manager.run()


from functools import wraps
from flask import jsonify, request

from silence.auth.tokens import check_token
from silence.server import manager as server_manager
from silence.db import dal
from silence import sql as SQL
from silence.sql import get_sql_op
from silence.sql.converter import silence_to_mysql
from silence.settings import settings
from silence.exceptions import EndpointError, HTTPError, TokenError
from silence.logging.default_logger import logger
from silence.utils.min_type import Min

import inspect
import re

OP_VERBS = {
    SQL.SELECT: 'get',
    SQL.INSERT: 'post',
    SQL.UPDATE: 'put',
    SQL.DELETE: 'delete',
}

RE_QUERY_PARAM = re.compile(r"^.*\$\w+/?$")

###############################################################################
# This is where the fun at
###############################################################################

def endpoint(route, method, sql, auth_required=False, description=None):
    logger.debug(f"Setting up endpoint {method} {route}")

    # Construct the API route taking the prefix into account
    route_prefix = settings.API_PREFIX
    if route_prefix.endswith("/"):
        route_prefix = route_prefix[:-1]  # Drop the final /
    full_route = route_prefix + route

    # Warn if the pair SQL operation - HTTP verb is not the proper one
    check_method(sql, method, route)

    # Extract the list of parameters that the user expects to receive
    # in the URL and in the SQL string
    sql_params = extract_params(sql)
    url_params = extract_params(route)

    # Get the required SQL operation
    sql_op = get_sql_op(sql)

    # If it's a SELECT or a DELETE, make sure that all SQL params can be
    # obtained from the url
    if sql_op in (SQL.SELECT, SQL.DELETE):
        check_params_match(sql_params, url_params, route)

    # wrapper is the inner decorator, so that the outer one can receive parameters
    def wrapper(func):
        @wraps(func)
        # The function itself that is being decorated (func)
        def decorator(*args, **kwargs):
            func(*args, **kwargs)

        # Get the list of arguments expected by the decorated function
        decorated_func_args = inspect.getargspec(func)[0]

        # If it's a SELECT or a DELETE, make sure that all SQL params can be
        # obtained from the url AND the request body
        if sql_op in (SQL.INSERT, SQL.UPDATE):
            check_params_match(sql_params, url_params + decorated_func_args, route)

        # The handler function that will be passed to flask
        def route_handler(*args, **kwargs):
            # If this endpoint requires authentication, check that the
            # user has provided a session token and that it is valid
            if auth_required:
                check_session()

            # Collect all url pattern params
            request_params = kwargs
            url_pattern_params = tuple(request_params[param] for param in url_params)

            # Convert the silence-style placeholders in the SQL query to proper MySQL placeholders
            query_string = silence_to_mysql(sql)

            # Default outputs
            res = None
            status = 200
            
            # SELECT/GET operations
            if sql_op == SQL.SELECT:
                # Call the decorated function, just in case (though it should do nothing)
                decorator()

                # The URL params have been checked to be enough to fill all SQL params
                res = dal.api_safe_query(query_string, url_pattern_params)
                
                # Filter these results according to the URL query string, if there is one
                # Possible TO-DO: do this by directly editing the SQL query for extra efficiency
                res = filter_query_results(res, request.args)

                # In our teaching context, it is safe to assume that if the URL ends
                # with a parameter and we have no results, we should return a 404 code
                if RE_QUERY_PARAM.match(route) and not res:
                    raise HTTPError(404, "Not found")

            else:  # POST/PUT/DELETE operations
                # Construct a dict for all params expected in the request body,
                # setting them to None if they have not been provided
                form = request.json if request.is_json else request.form
                body_params = {param: form.get(param, None) for param in decorated_func_args}

                # Call the decorated function with these parameters to allow the
                # user to validate them
                decorator(**body_params)

                # We have checked that sql_params is a subset of url_params U body_params,
                # construct a joint param object and use it to fill the SQL placeholders
                for i, param in enumerate(url_params):
                    body_params[param] = url_pattern_params[i]

                param_tuple = tuple(body_params[param] for param in sql_params)

                # Run the execute query
                res = dal.api_safe_update(query_string, param_tuple)

            return jsonify(res), status
        
        # flaskify_url() adapts the URL so that all $variables are converted to Flask-style <variables>
        server_manager.APP.add_url_rule(flaskify_url(full_route), method + route, route_handler, methods=[method])
        server_manager.API_TREE.add_url(full_route)
        server_manager.API_TREE.register_endpoint({"route": full_route, "method": method.upper(), "description": description})

        return decorator
    return wrapper

###############################################################################
# Session token checker
def check_session():
    token = request.headers.get("Token", default=None)
    if not token:
        raise HTTPError(401, "Unauthorized")

    try:
        check_token(token)
    except TokenError as exc:
        raise HTTPError(401, str(exc))

###############################################################################
# Aux stuff for the handler function

# Implements filtering, ordering and paging using query strings
def filter_query_results(data, args):
    # Grab all parameters from the query string
    sort_param = args.get("_sort", None)
    sort_reverse = "_order" in args and args["_order"] == "desc"

    try:
        limit = int(args.get("_limit", None))
    except (ValueError, TypeError):
        limit = None

    try:
        page = int(args.get("_page", None))
    except (ValueError, TypeError):
        page = None

    # Filter, sort and paginate results
    filter_criteria = [pair for pair in args.items() if not pair[0][0] == "_"]
    filter_func = lambda elem: all(k not in elem or str(elem[k]).lower() == v.lower() for k, v in filter_criteria)
    res = list(filter(filter_func, data))

    def order_key_func(elem):
        v = elem[sort_param]
        return v if v is not None else Min  # Avoids comparisons against None

    try:
        res.sort(key=order_key_func, reverse=sort_reverse)
    except KeyError: pass

    offset = limit * page if limit and page else 0
    top = offset + limit if limit else len(res)
    return res[offset:top]

###############################################################################
# Aux stuff for checking and parsing

# Checks whether the SQL operation and the HTTP verb match
def check_method(sql, verb, endpoint):
    sql_op = get_sql_op(sql)

    if sql_op in OP_VERBS:
        correct_verb = OP_VERBS[sql_op]

        if correct_verb != verb.lower():
            # Warn the user about the correct verb to use
            logger.warn(
                f"The '{verb.upper()}' HTTP verb is not correct for the SQL {sql_op.upper()} " +
                f"operation in endpoint {verb.upper()} {endpoint}, the correct verb is {correct_verb.upper()}."
            )
    else:
        # What has the user put here?
        raise EndpointError(f"The SQL query '{sql}' in the endpoint {endpoint} is not supported," +
                             " please use only SELECT/INSERT/UPDATE/DELETE.")

# Returns a list of $params in a SQL query or endpoint route,
# without the $'s
def extract_params(string):
    res = re.findall(r"\$\w+", string)
    return [x[1:] for x in res]

# Convers $url_param to <url_param>
def flaskify_url(url):
    return re.sub(r"\$(\w+)", r"<\1>", url)

# Checks whether all SQL params can be filled with the provided params
def check_params_match(sql_params, user_params, route):
    sql_params = set(sql_params)
    user_params = set(user_params)
    diff = sql_params.difference(user_params)

    if diff:
        params_str = ", ".join(f"${param}" for param in diff)
        raise EndpointError(f"Error creating endpoint {route}: the parameters " +
                            f"{params_str} are expected by the SQL query but they are not provided in the URL " +
                            "or the request body.")